open Cil
open List
open Pretty

module E = Errormsg

(*
 * As we recurse into the AST, we need to pass along some state about
 * the boolean variable governing the `if' block we're in, as well as
 * information about the loop we're in, if any.  Furthermore, we need
 * to be able to simulate returns from functions independent of any
 * conditionals or loops we are in.
 *)
type xstate =
    { mutable ifcond   : exp;
      mutable loopcond : exp;
      mutable retcond  : exp;
    }

let copyXs (xs : xstate) : xstate =
  {ifcond=xs.ifcond; loopcond=xs.loopcond; retcond=xs.retcond}

let cBinOp (b, e1, e2, t) =
  let op = BinOp(b, e1, e2, t) in
    if !Cil.foldingOn then constFold false op else op

let cUnOp (b, e, t) =
  let op = UnOp(b, e, t) in
    if !Cil.foldingOn then constFold false op else op

let dummyVisitor = new nopCilVisitor

let iterProto = ref []

(*
 * Examine a function prototype, looking for our magic attribute.
 * A function that is declared with this attribute becomes a
 * magic marker signifying the maximum number of iterations a `for'
 * loop can make.  We use this kludge because __attribute__ tags
 * can only be applied to declarations and function definitions.
 *)
let powerProto (f: Cil.varinfo) : unit =
  if filterAttributes "imarker" f.vattr <> [] then
    iterProto := [f]

(*
 * CIL very annoyingly combines adjacent instructions into a sublist of
 * instructions.  Rather than do something clever, we just undo this feature.
 * XXX We really should be using the builtin visitor pattern routines,
 * or we're sure to have missed some types of statements that have sub-trees.
 *)
let unroll (f: Cil.fundec) : unit =
  let rec
  processStmts = function
    | [] -> []
    | ({skind=Instr(f :: r)} as s) :: rests when r <> [] ->
        mkStmtOneInstr f ::
        processStmts [mkStmt (Instr r)] @
        processStmts rests
    | ({skind=If (ifPredicate, ifConseq, ifAlt, l)} as s) :: rests ->
        let newConseq = processBlock ifConseq and
            newAlt = processBlock ifAlt in
          s.skind <- If (ifPredicate, newConseq, newAlt, l);
          s :: processStmts rests
    | ({skind=Loop(blk,l,opt1,opt2)} as s) :: rests ->
        let newBlk = processBlock blk in
          s.skind <- Loop(newBlk, l, opt1, opt2);
          s :: processStmts rests
    | s :: rests ->
        s :: processStmts rests
  and processBlock (b: block) =
    { bstmts = processStmts b.bstmts; battrs = b.battrs; }
  in
  f.sbody <- processBlock f.sbody

(* The Big Kahuna: Scan a function definition *)
let power (f: Cil.fundec) : unit = 
  let fname = f.svar.vname in
  let retTyp = 
    match f.svar.vtype with
      TFun(rt, _, _, _) -> rt
    | _ -> E.s (E.bug "Function %s does not have a function type\n" 
                  f.svar.vname) in
  let
  retVar : varinfo option ref =
    match retTyp with
      TVoid _ -> ref None
    | _ -> ref (Some (makeLocalVar f "__retval" retTyp))
  and
  (* Make a new flag variable to store the result of a predicate *)
  mkFlagVar () : lhost =
    Var(makeTempVar f ~name:("__c") charType)
  and
  mkLoopVar () : varinfo =
    makeTempVar f ~name:("__f") intType
  and
  (*
   * Convert a 0 or 1 value in any type to a mask of all 0's or all 1's
   * in the desired target type.
   *)
  condToMask (conds : xstate) (ttype : typ) : exp =
    let c = cBinOp(BAnd, conds.ifcond, conds.loopcond, charType) in
    let d = cBinOp(BAnd, conds.retcond, c, charType) in
      cBinOp(MinusA, d, one, ttype)
  and
  loopTag =
    match !iterProto with
      proto :: rest -> proto.vname
    | [] -> ""
  in

  let rec
  (*
   * Using arithmetic tricks, construct a statement semantically
   * equivalent to `lval = cond ? r1 : r2'.
   *)
  condAssign (left: lval) (r1: exp) (r2: exp) loc conds : stmt list =
    let ltype = typeOf(Lval left) in
    let mask = condToMask conds ltype in
    let nmask = cUnOp(BNot, mask, ltype) in
    let ifTrue = cBinOp(BAnd, nmask, r1, ltype) and
        ifFalse = cBinOp(BAnd, mask, r2, ltype) in
    let newRval = cBinOp(BOr, ifTrue, ifFalse, ltype) in
      [mkStmt(Instr [Set(left, newRval, loc)])]
  and
  (* Construct `if (cond) left = right;' *)
  rewriteAssign ({skind=Instr[Set(left, right, loc)]} as s) conds : stmt list =
    condAssign left right (Lval left) loc conds
  and
  (*
   * Rewrite an `if' statement.
   * Note that `var' is a new variable that holds the result of this
   * `if' predicate, whereas `cond' is the condition under which the
   * entire `if' block is or isn't executed.
   *)
  rewriteIf ({skind = If(ifPred, ifCons, ifAlt, l)} as s) conds : stmt list =
    (* [l]var is used to hold the result of the if predicate *)
    let var = mkFlagVar() in
    let lvar = Lval(var, NoOffset) in
    let notlvar = cBinOp(MinusA, one, lvar, intType) and (* inverse of var *)
	(* Construct a version of the predicate that is always 0 or 1 *)
	boolPred = cBinOp(Gt, ifPred, zero, typeOf(ifPred)) in
    (*
     * Create an assignment of var to the result of the predicate,
     * but only if we're supposed to execute this entire `if' statement
     * at all.  If not, the var is automatically false.
     *)
    let cAssign = condAssign (var,NoOffset) boolPred zero l conds and
        newcond = copyXs conds and
        newnotcond = copyXs conds in
      newcond.ifcond <- lvar;
      newnotcond.ifcond <- notlvar;
    (* Construct blocks for the two branches of the `if' *)
    let newCons = mkStmt(Block(scanBlock newcond ifCons)) and
        newAlt = mkStmt(Block(scanBlock newnotcond ifAlt)) in
      cAssign @ [newCons; newAlt]
  and
  rewriteLoop (nIters : exp) (conds: xstate) (blk : block) (loc : location)
    : stmt list =
    let fvar = mkLoopVar() and
        cvar = mkFlagVar() in   (* indicates that we're not done looping *)
    let ccond = copyXs conds in
      ccond.loopcond <- (Lval(cvar,NoOffset));
    let cAssign = condAssign (cvar,NoOffset) one zero loc conds and
        xformedBody = mkStmt(Block(scanBlock ccond blk)) in
      (* XXX The library routine mkForIncr doesn't work with `continue' *)
      cAssign @ mkForIncr fvar zero nIters one [xformedBody]
  and
  rewriteBreak (conds: xstate) (loc: location) : stmt list =
    let left = match conds.loopcond with
      Lval(lv) -> lv
    | _ -> failwith "loopcond not an lvalue" in
    let cAssign = condAssign left zero conds.loopcond loc conds in
      cAssign
  and
  rewriteReturn (conds: xstate) (loc: location) : stmt list =
    let left = match conds.retcond with
      Lval(lv) -> lv
    | _ -> failwith "retcond not an lvalue" in
      condAssign left zero conds.retcond loc conds
  and
  scanStmts (conds: xstate) = function
    | [] -> []
    | ({skind=If (ifPredicate, ifConseq, ifAlt, l)} as s) :: rests ->
        rewriteIf s conds @ scanStmts conds rests
    | {skind=Instr[Call(_,Lval(Var(fn),NoOffset),args,l)]} as s :: rests
      when fn.vname=loopTag ->
        begin
        match args with
          Const(cval) :: restargs ->
            begin
            match rests with
              {skind=Loop(blk,l,_,_)} :: rest1 ->
                  s.skind <- Instr[];
                  s :: rewriteLoop (Const(cval)) conds blk l
                    @ scanStmts conds rest1
              | s :: rest1 ->
                  ignore(E.log " %a\n" d_stmt (s));
                  failwith "misplaced loop iteration annotation"
              | [] ->
                  failwith "loop iteration annotation at end of function"
            end
        | _ ->
            ignore(E.log " %a\n" d_stmt (s));
            failwith "bad args to iteration annotation"
        end
    | ({skind=Loop(_,_,_,_)} as s) :: rests ->
       (*
        * All loops should be immediately preceded by an annotation
        * and handled earlier, so we should never get here.
        *)
        ignore(E.log " %a\n" d_stmt (s));
        failwith "unannotated loop"
    | ({skind=Break(l)} as s) :: rests ->
        rewriteBreak conds l @ scanStmts conds rests
    | ({skind=Return (retval, l)} as s) :: rests ->
        let newret = rewriteReturn conds l in
        let retAssign = match retval with
              Some rval ->
                begin
                match !retVar with
                  Some rvar ->
                    condAssign (Var(rvar),NoOffset) rval
                      (Lval(Var(rvar),NoOffset)) l conds
                    @ newret
                | None ->
                    failwith "Return value in void function"
                end
            | None -> newret in
          retAssign @ scanStmts conds rests
    | ({skind=Instr[Set(left, right, l)]} as s) :: rests ->
        rewriteAssign s conds @ scanStmts conds rests
    | s :: rests ->
        s :: scanStmts conds rests
  and scanBlock (conds: xstate) (b: block) =
    { bstmts = scanStmts conds b.bstmts; battrs = b.battrs; }

  in
  let retcond = (Var(makeLocalVar f "__ret" charType),NoOffset) in
  let rvar = match !retVar with
      Some rv -> Some (Lval(Var(rv), NoOffset))
    | None -> None in
  let retcondAssign = mkStmt(Instr [Set(retcond, one, f.svar.vdecl)]) and
      realRet = mkStmt(Return(rvar, locUnknown))
  in
    ignore (visitCilBlock dummyVisitor f.sbody) ; (* sets CurrentLoc *)
    f.sbody <-
      scanBlock {ifcond=one; loopcond=one; retcond=Lval(retcond)} f.sbody;
    f.sbody <-
      { bstmts = retcondAssign :: f.sbody.bstmts @ [realRet];
        battrs = f.sbody.battrs;
      }

let feature : featureDescr =
  { fd_name = "power";
    fd_enabled = Cilutil.doPower;
    fd_description = "Transform out power and timing vulnerabilities" ;
    fd_extraopt = [];
    fd_doit = (function (f: file) -> 
      (***
      if not !Cilutil.doPartial then begin
        Errormsg.s (Errormsg.error
             "--dopower: you must also specify --dopartial --domakeCFG\n")
      end ;
      ***)
      Cil.iterGlobals f (fun glob -> match glob with
        Cil.GVarDecl(proto, _) when isFunctionType proto.vtype ->
          powerProto proto
      | Cil.GFun(fd,_) ->
        unroll fd;
        power fd;
      | _ -> ()));
    fd_post_check = true;
  } 
