import java.io.*;
import java.util.*;
import java.util.regex.*;

/**
 * Instances of this class represent nodes in the control flow graph
 * of the program.
 */
class AsmNode {
    /** The set of all labels in the program, shared among all AsmNodes */
    protected Map<String, AsmNode> labels;

    /** All call instructions in the program */
    protected Set<AsmNode> callSites;

    /** All external (.globl) variables */
    protected Map<String, AsmNode> globals;

    /** The label immediately preceding this node, if applicable */
    // XXX Technically, there could be more than one.
    protected String myLabel;

    /** The global symbol that refers to this address. */
    // XXX There could be more than one; also, we don't handle .equ.
    protected String mySym;

    protected Insn instruction;

    /** Possible control paths from here */
    protected LinkedList<AsmNode> paths;

    /** The next instruction in the source file, to facilitate toString() */
    protected AsmNode nextNodeInFile;

    /**
     * Given a Reader that has an assembly program, this factory method
     * parses the program and constructs a CFG for it.  Note that the
     * node that is returned represents only the first instruction in
     * the program.  However, each node knows about all the labels in
     * the entire program, so any reachable code can be accessed through
     * one of the labels.
     * (This is a factory method instead of a constructor because it may
     * return null if there is no more input to be read from the Reader.)
     */
    public static AsmNode makeAsmNode(Reader in) {
	return makeAsmNode(new LineNumberReader(in),
			   new AsmNode(new HashMap<String, AsmNode>(256),
				       new HashSet<AsmNode>(),
				       new HashMap<String, AsmNode>(8)));
    }

    private AsmNode(Map<String, AsmNode> labels, Set<AsmNode> callSites,
		    Map<String, AsmNode> globals)
    {
        this.labels = labels;
	this.callSites = callSites;
	this.globals = globals;
        paths = new LinkedList<AsmNode>();
    }

    /**
     * This is a copy constructor for the benefit of makeAsmNode().
     * All AsmNodes corresponding to any given program share the same
     * labels and callSites.  (I didn't want to use static variables,
     * though, because that would make it impossible to handle multiple
     * source files later on.)
     */
    private AsmNode(AsmNode parent)
    {
	this.labels = parent.labels;
	this.callSites = parent.callSites;
	this.globals = parent.globals;
	paths = new LinkedList<AsmNode>();
    }

    /**
     * This is an internal factory method.  We call it recursively to
     * construct the list of AsmNodes representing the program.
     * (It isn't an ordinary constructor because it may return null if
     * there is no more input to be read from the Reader.)
     */
    protected static AsmNode makeAsmNode(LineNumberReader lines,
					 AsmNode parent)
    {
	AsmNode node = new AsmNode(parent);

	try {
	    /*
	     * We split the input into lines, which are in turn split
	     * into tokens.  The splitting characters are whitespace
	     * and commas, but we don't differentiate between the two.
	     * This may mean that we accept some programs that the
	     * system assembler wouldn't, but I don't think this
	     * introduces any ambiguitites.
	     *
	     * XXX We don't handle semicolons, e.g. ``op1 ; op2''
	     *
	     * XXX We don't handle relative labels, e.g. 1f, 1b
	     */
	    while (true) {
		String line = lines.readLine();
		if (line == null) {
		    // Oops, there's no instruction to be read (EOF).
		    // Make sure no labels or global symbols point to us.
		    if (node.myLabel != null)
			node.labels.remove(node.myLabel);
		    if (node.mySym != null)
			node.globals.remove(node.mySym);
		    return null;
		}

		// Trim comments
		int k = line.indexOf('#');
		if (k >= 0)
		    line = line.substring(0, k);

		String[] tokens = tokenize(line.trim());
		Debug.println(3, "tokens: " + Arrays.asList(tokens));

		// Blank lines: ignore
		if (tokens.length == 0 || tokens[0].equals(""))
		    continue;

		// Labels
		if (tokens[0].endsWith(":")) {
		    String label = tokens[0];
		    label = label.substring(0, label.length() - 1);
		    Debug.println(3, "Adding label " + label);
		    node.labels.put(label, node);
		    node.myLabel = label;
		    // An opcode might be on the same line as the label
		    if (tokens.length > 1) {
			// Could probably be done better...
			String[] newtokens = new String[tokens.length - 1];
			System.arraycopy(tokens, 1, newtokens, 0,
					 newtokens.length);
			tokens = newtokens;
		    } else {
			continue;
		    }
		}

		// Assembler directives
		if (tokens[0].startsWith(".")) {
		    if (tokens[0].equals(".globl")) {
			// Global symbol
			// XXX Some global symbols might refer to data and
			//     not code.  In these cases, we will assume that
			//     the next instruction is the target anyway.
			if (tokens.length != 2) {
			    throw new BadInsnException(lines.getLineNumber(),
					   "wrong number of args to .globl");
			}
			node.mySym = tokens[1];
			Debug.println(3, "Adding globl " + node.mySym);
			node.globals.put(node.mySym, node);
			continue;
		    } else {
			// Something else: ignore
			continue;
		    }
		}

		// Opcode, presumably
		node.instruction = InsnIA32.makeInsn(tokens,
						     lines.getLineNumber());
		Debug.println(3, "Adding op: " + node.instruction.getOpcode());
		/*
		 * Recursively process the rest of the lines.
		 * After we do this, we will also have learned about
		 * all of the labels in the program.  ORDER IS CRUCIAL HERE,
		 * since instantiating AsmNodes for the rest of the program
		 * ensures that we know about all the labels and call sites
		 * in the program.
		 *
		 * NB: We make the next instruction a successor of this one
		 *     unless this one is an unconditional jump.  This logic
		 *     needs to change on architectures that support
		 *     delayed branches.
		 */
		node.nextNodeInFile = makeAsmNode(lines, node);
		if (!(node.instruction.isUnconditionalControl()) &&
		    node.nextNodeInFile != null)
		    node.paths.add(node.nextNodeInFile);
		// If this is a branch, add the appropriate CFG edge.
		if (node.instruction.hasLabelTarget()) {
		    AsmNode target =
			node.labels.get(node.instruction.getLabel());
		    if (target == null) {
			throw new BadInsnException("unknown label " +
						  node.instruction.getLabel());
		    }
		    node.paths.add(node.labels.get(
				       node.instruction.getLabel()));
		} else if (node.instruction.isUnconditionalControl()) {
		    if (node.instruction.isCall()) {
			if (node.nextNodeInFile != null)
			    node.callSites.add(node.nextNodeInFile);
		    } else if (node.instruction.isReturn()) {
			/*
			 * Rather than trying to figure out what set of
			 * labels we might possibly need to return to,
			 * we assume that we could return to any call site.
			 */
			node.paths.addAll(node.callSites);
		    }
		}
		break;
	    }
	    return (node);
	} catch (IOException e) {
	    System.err.println("IOException: " + e);
	} catch (ArrayIndexOutOfBoundsException e) {
	    // XXX This is a hack.  We should actually check the array
	    // bounds more carefully, rather than assuming that an
	    // ArrayOutOfBoundsException means a syntax error occurred.
	    System.err.println("Syntax error, line " + lines.getLineNumber());
	    System.err.println("Expecting more tokens");
	} catch (BadInsnException e) {
	    System.err.println("Line " + lines.getLineNumber() + ": " + e);
	}
	return (null);
    }

    public List<AsmNode> getPaths() {
	return (Collections.unmodifiableList(paths));
    }

    /**
     * ``Execute'' this instruction.  The argument is the set of data values
     * that are untainted.  The same object is returned, but with newly-
     * tainted values removed and newly-untainted values added.
     */
    public LocationSet execute(Configuration c) {
	return instruction.execute(c);
    }

    public Map<String, AsmNode> getGlobals() {
	return Collections.unmodifiableMap(globals);
    }

    public String toString() {
	List<String> ls = toStringList();
	StringBuffer sb = new StringBuffer(ls.size() * 40);
	for (String elt : ls)
	    sb.append(elt);
	return sb.toString();
    }

    protected String myString() {
	String s = "";
        if (mySym != null)
            s += ".globl " + mySym + "\n";
        if (myLabel != null)
            s += myLabel + ":\n";
        return s + instruction.toString() + "\n";
    }

    protected LinkedList<String> toStringList() {
	LinkedList<String> rest;
	if (nextNodeInFile == null)
	    rest = new LinkedList<String>();
	else
	    rest = nextNodeInFile.toStringList();
	rest.addFirst(myString());
	return (rest);
    }

    /**
     * Java's regexp engine is somewhat limited, so we can't make
     * String.split() understand that a comma that appears in a single
     * level of parentheses isn't a delimiter.  (This is possible in
     * regular languages in general because the parentheses are never
     * nested more than one level.)  What we do instead is let split()
     * screw up, then coalesce any strings that it got wrong.
     * The cases it gets wrong are things like 'leal (%eax,%ecx), %eax',
     * where the '(%eax,%ecx)' is supposed to be a single operand.
     *
     * XXX This is a hack.
     */
    protected static String[] tokenize(String line) {
	String[] result = line.split("[\\s,]+");
	for (int i = 0; i < result.length - 1; i++) {
	    if (Pattern.matches("[(][^)]+\\z", result[i])) {
		String[] newResult = new String[result.length - 1];
		System.arraycopy(result, 0, newResult, 0, i);
		newResult[i] = result[i] + "," + result[i+1];
		System.arraycopy(result, i + 2, newResult, i + 1,
				 newResult.length - i - 1);
		result = newResult;
	    }
	}
	return result;
    }

}
