import java.util.*;

/**
 * Instances of this class represent configurations of an abstract
 * machine.  The machine models an x86 processor, but we don't care
 * about actual register values or directions of jumps.  Instead, we
 * only care whether data are tainted (i.e. they depend on secret
 * inputs) or untainted, and we care that branches are made based
 * only on untainted data.
 */
public class Configuration {
    protected static Set<Configuration> allConfigs;
    
    AsmNode pc;
    protected LocationSet untaintedData;

    public Configuration(AsmNode pc) {
	this.pc = pc;
	untaintedData = new LocationSet(32);
    }

    protected Configuration(AsmNode pc, LocationSet untaintedData)
    {
	this.pc = pc;
	this.untaintedData = untaintedData;
    }

    public void simulate() {
	allConfigs = new HashSet<Configuration>();
	simulateInternal();
    }

    protected void simulateInternal() {
	allConfigs.add(this);

	List<AsmNode> paths = pc.getPaths();
	LocationSet nextdata = pc.execute(this);
	for (AsmNode nextpc : paths) {
	    Configuration nextConfig = new Configuration(nextpc, nextdata);
	    if (!allConfigs.contains(nextConfig))
		nextConfig.simulateInternal();
	}
    }

    /**
     * Return the untainted data in this configuration.  The caller may
     * NOT modify the set, although this restriction is not enforced
     * because Java's generics implementation makes this difficult.
     */
    public LocationSet getUntaintedData() {
	return untaintedData;
    }

    public int hashCode() {
	return pc.hashCode() + untaintedData.hashCode();
    }

    public boolean equals(Object other) {
	if (!(other instanceof Configuration))
	    return false;
	Configuration oc = (Configuration)other;
	return (pc.equals(oc.pc) && untaintedData.equals(oc.untaintedData));
    }
}
