import java.util.*;
import java.util.regex.*;

/**
 * Instances of this class represent assembly operands:
 * constants, registers, and memory locations.
 *
 * TODO:
 * - Understand that %eax, %ax, %ah, and %al all refer to parts of
 *   the same register.
 */
public class DataValue {
    protected String name;
    protected locType type;

    /**
     * The set of registers that this operand depends on.
     * A value d1 depends on d2 if the *identity* of the location named
     * by d1 is defined in terms of d2.  For instance, %(eax) depends on
     * %eax.  Once instantiated, this set is immutable.
     */
    protected Set<DataValue> regDeps = Collections.emptySet();

    public static enum locType {
	REG, MEM, CONST;
    }

    public DataValue(String name) throws BadInsnException {
	this.name = name;
	type = classify();
	if (type == locType.MEM)
	    setMemDeps();
	Debug.println(3, "DEPS of " + name + " = " + regDeps);
    }

    /**
     * Parse the operand and classify it as REG, MEM, or CONST.
     */
    protected locType classify() throws BadInsnException {
	if (Pattern.matches("\\A\\$-?(0[Xx])?[0-9]+\\z", name))
	    return locType.CONST;			 // e.g. $0x12
	else if (Pattern.matches("\\A%[a-z]+\\z", name))
	    return locType.REG;				 // e.g. %eax
	else if (name.equals("eflags"))
	    return locType.REG;
	else if (Pattern.matches("\\A(-?[0-9]+)?\\(%[a-z]+\\)\\z", name))
	    return locType.MEM;				 // e.g. -12(%eax)
	else if (Pattern.matches("\\A\\(%[a-z]+,%[a-z]+\\)\\z", name))
	    return locType.MEM;				 // e.g. (%eax,%ecx)
	else if (Pattern.matches("\\A[a-z]+\\z", name))
	    return locType.MEM;				 // e.g. foo
	else
	    throw new BadInsnException("can't parse operand " + name);
    }

    // Base+offset form of pointer using 1 register
    private static final Pattern memDispPat =
	Pattern.compile("\\A(?:-?[0-9]+)?\\((%[a-z]+)\\)\\z");
    // Scaled base + n*offset form pointer using 2 registers
    private static final Pattern memDispScaledPat =
	Pattern.compile("\\A\\((%[a-z]+),(%[a-z]+)\\)\\z");
    // Constant pointer
    private static final Pattern memConstPat =
	Pattern.compile("\\A[a-z]+\\z");
    
    protected void setMemDeps() {
	assert(type == locType.MEM);
	Matcher m = memDispPat.matcher(name);
	DataValue d;
	try {
	    if (m.matches()) {
		regDeps = Collections.singleton(new DataValue(m.group(1)));
		return;
	    }
	    m = memDispScaledPat.matcher(name);
	    if (m.matches()) {
		regDeps = new HashSet<DataValue>(2);
		regDeps.add(new DataValue(m.group(1)));
		regDeps.add(new DataValue(m.group(2)));
		regDeps = Collections.unmodifiableSet(regDeps);
		return;
	    }
	    m = memConstPat.matcher(name);
	    if (m.matches())
		return;
	} catch (BadInsnException e) {
	    throw new AssertionError("caught BadInsnException while " +
				     "computing dependencies: " + name);
	}
	throw new AssertionError("can't classify memory operand");
    }

    public locType getType() {
	return type;
    }

    public Set<DataValue> getDependencies() {
	return regDeps;
    }

    public String toString() {
	return name;
    }

    public boolean equals(Object other) {
	if (!(other instanceof DataValue))
	    return false;
	else
	    return name.equals(((DataValue)other).name);
    }

    public int hashCode() {
	return name.hashCode();
    }
}
