import java.util.*;

/**
 * This class represents a single instruction (opcode and operands).
 * This is relatively low-level; we store the arguments as strings,
 * and there are no references to previous/subsequent instructions,
 * jump targets, etc.
 *
 * @see AsmNode
 */
abstract class Insn {
    protected String opcode;
    protected String label;	// XXX should probably be an object
    protected List<DataValue> inputs;
    protected List<DataValue> outputs;
    protected int lineNum;

    protected Insn(String tokens[]) throws BadInsnException {
	opcode = tokens[0];
    }

    public String getOpcode() {
	return opcode;
    }

    // XXX inheritance
    public String getLabel() throws BadInsnException {
	if (label != null)
	    return label;
	else
	    throw new BadInsnException("label requested, but not a jump");
    }

    public List<DataValue> getInputs() throws BadInsnException {
	if (inputs != null)
	    return Collections.unmodifiableList(inputs);
	else
	    throw new BadInsnException("getInputs(): wrong insn type");
    }

    public List<DataValue> getOutputs() throws BadInsnException {
	if (outputs != null)
	    return Collections.unmodifiableList(outputs);
	else
	    throw new BadInsnException("getOutputs(): wrong insn type");
    }

    public boolean isCall() {
	return false;
    }

    public boolean isReturn() {
	return false;
    }

    /**
     * Is this an unconditional control transfer (e.g. jmp, ret)?
     */
    public boolean isUnconditionalControl() {
	return false;
    }

    /**
     * Does this instruction have a label target (e.g. jump, call, branch)?
     */
    public boolean hasLabelTarget() {
	return false;
    }

    /**
     * ``Execute'' this instruction.  The argument is the set of data values
     * that are untainted.  Newly-tainted values are removed from the set and
     * newly-untainted values are added.
     */
    public abstract LocationSet execute(Configuration c);

    // Is item tainted with respect to the given dataset?
    protected final boolean isTainted(LocationSet data, DataValue item) {
	if (item.getType() == DataValue.locType.CONST)
	    return false;
	else
	    return !data.contains(item);
    }

    /**
     * It's impossible to know whether, say, (%eax) and (%ecx) alias
     * each other in general.  We adopt a conservative model, in which
     * we keep track of the last memory reference, and any time we see
     * a reference through a different register, we taint all memory
     * locations.
     */
    protected void checkMemoryAliasing(LocationSet data) {
	Set<DataValue> lastOpDeps;
	if (data.getLastMemOp() == null)
	    lastOpDeps = Collections.emptySet();
	else
	    lastOpDeps = data.getLastMemOp().getDependencies();

	for (DataValue in : inputs) {
	    if (in.type != DataValue.locType.MEM)
		continue;
	    if (!in.getDependencies().equals(lastOpDeps)) {
		data.flushMemory();
		data.setLastMemOp(in);
		lastOpDeps = in.getDependencies();
	    }
	}
	for (DataValue out : outputs) {
	    if (out.type != DataValue.locType.MEM)
		continue;
	    if (!out.getDependencies().equals(lastOpDeps)) {
		data.flushMemory();
		data.setLastMemOp(out);
		lastOpDeps = out.getDependencies();
	    }
	}
    }


    /**
     * Anything with inputs and outputs
     */
    static abstract class OpAssign extends Insn {
	OpAssign(String tokens[]) throws BadInsnException {
	    super(tokens);
	    inputs = new LinkedList<DataValue>();
	    outputs = new LinkedList<DataValue>();
	}

	public String toString() {
	    return lineNum + "\t" + opcode +
		"\tin: " + inputs + " out: " + outputs;
	}

	public LocationSet execute(Configuration c) {
	    LocationSet data = (LocationSet)c.getUntaintedData().clone();

	    checkMemoryAliasing(data);

	    // If any input is tainted, taint the output.
	    boolean taintOutputs = false;
	    for (DataValue in : inputs) {
		if (isTainted(data, in)) {
		    taintOutputs = true;
		    break;
		}
	    }
	    if (taintOutputs) {
		for (DataValue out : outputs)
		    data.remove(out);
	    } else {
		for (DataValue out : outputs)
		    data.add(out);
	    }

	    /*
	     * If an output operand was used to name a memory location,
	     * we need to re-taint any memory locations that it might
	     * have referred to.  This is true even if the new value
	     * being assigned to the pointer is a constant, since we know
	     * nothing about the new target of the pointer.
	     *
	     * XXX This is a bit expensive.
	     */
	    LinkedList<DataValue> toBeRemoved = new LinkedList<DataValue>();
	    for (DataValue d : data) {
		for (DataValue out : outputs) {
		    if (d.getDependencies().contains(out))
			toBeRemoved.add(d);
		}
	    }
	    for (DataValue d : toBeRemoved)
		data.remove(d);

	    return data;
	}
    }

    /**
     * Anything that transfers control
     */
    static abstract class OpControl extends Insn {
	OpControl(String tokens[]) throws BadInsnException {
	    super(tokens);
	}
    }

    /**
     * Unconditional control transfer, i.e. jmp and ret
     */
    static abstract class OpUCControl extends OpControl {
	OpUCControl(String tokens[]) throws BadInsnException {
	    super(tokens);
	}

	/** For unconditional control transfer, there's nothing to check. */
	public LocationSet execute(Configuration c) {
	    return c.getUntaintedData();
	}
    }

    /**
     * Anything that modifies the stack pointer
     */
    static abstract class OpStack extends OpAssign {
	OpStack(String tokens[]) throws BadInsnException {
	    super(tokens);
	    inputs.add(new DataValue("%esp"));
	    outputs.add(new DataValue("%esp"));
	}
    }

    /**
     * An assignment of the form OP IO, where operand and IO is both an
     * input and output operand.  Modifies eflags.
     */
    static class OpAssignIO1 extends OpAssign {
	OpAssignIO1(String tokens[]) throws BadInsnException {
	    super(tokens);
	    assert(amI(tokens[0]));
	    if (tokens.length != 2) {
		throw new BadInsnException("wrong number of parameters: " +
					   tokens[0]);
	    }
	    inputs.add(new DataValue(tokens[1]));
	    outputs.add(new DataValue(tokens[1]));
	    outputs.add(new DataValue("eflags"));
	}

	final private static Collection names =
	    Arrays.asList("decl", "decw", "decb",
			  "incl", "incw", "incb",
			  "negl", "negw", "negb");

	static boolean amI(String op) {
	    return names.contains(op);
	}
    }

    /**
     * An assignment of the form OP IO, where operand and IO is both an
     * input and output operand.  Does not modify eflags.
     */
    static class OpAssignIO1NF extends OpAssign {
	OpAssignIO1NF(String tokens[]) throws BadInsnException {
	    super(tokens);
	    assert(amI(tokens[0]));
	    if (tokens.length != 2) {
		throw new BadInsnException("wrong number of parameters: " +
					   tokens[0]);
	    }
	    inputs.add(new DataValue(tokens[1]));
	    outputs.add(new DataValue(tokens[1]));
	}

	final private static Collection names =
	    Arrays.asList("notl", "notw", "notb");

	static boolean amI(String op) {
	    return names.contains(op);
	}
    }

    /**
     * An assignment of the form OP IN, IO, where IN is an input
     * operand and IO is both an input and output operand.
     */
    static class OpAssignIO2 extends OpAssign {
	OpAssignIO2(String tokens[]) throws BadInsnException {
	    super(tokens);
	    assert(amI(tokens[0]));
	    if (tokens.length != 3) {
		throw new BadInsnException("wrong number of parameters: " +
					   tokens[0]);
	    }
	    inputs.add(new DataValue(tokens[1]));
	    inputs.add(new DataValue(tokens[2]));
	    outputs.add(new DataValue(tokens[2]));
	    outputs.add(new DataValue("eflags"));
	}

	/*
	 * XXX I assume that all of these set all 
	 * the interesting condition flags.  Only mostly true.
	 */
	final private static Collection names =
	    Arrays.asList("addl", "addw", "addb",
			  "andl", "andw", "andb",
			  "imull", "imulw", "imulb", // can also have 3 args
			  "orl",  "orw",  "orb",
			  "sall", "salw", "salb",
			  "sarl", "sarw", "sarb",
			  "sbbl", "sbbw", "sbbb",
			  "shll", "shlw", "shlb",
			  "shrl", "shlw", "shlb",
			  "subl", "subw", "subb",
			  "xorl", "xorw", "xorb");

	static boolean amI(String op) {
	    return names.contains(op);
	}
    }

    /**
     * Identical to OpAssignIO2, but eflags is also an input.
     */
    static class OpAssignIO2F extends OpAssignIO2 {
	OpAssignIO2F(String tokens[]) throws BadInsnException {
	    super(tokens);
	    assert(amI(tokens[0]));
	    inputs.add(new DataValue("eflags"));
	}

	final private static Collection names =
	    Arrays.asList("adcl", "adcw", "adcb");

	static boolean amI(String op) {
	    return names.contains(op);
	}
    }

    /**
     * Compare and set flags
     */
    static class OpCmp extends OpAssign {
	OpCmp(String tokens[]) throws BadInsnException {
	    super(tokens);
	    assert(amI(tokens[0]));
	    if (tokens.length != 3) {
		throw new BadInsnException("wrong number of parameters: " +
					   tokens[0]);
	    }
	    inputs.add(new DataValue(tokens[1]));
	    inputs.add(new DataValue(tokens[2]));
	    outputs.add(new DataValue("eflags"));
	}

	final private static Collection names =
	    Arrays.asList("cmpl", "cmpw", "cmpb",
			  "testl", "testw", "testb");

	static boolean amI(String op) {
	    return names.contains(op);
	}
    }

    /**
     * Set on condition code
     */
    static class OpSetCC extends OpAssign {
	OpSetCC(String tokens[]) throws BadInsnException {
	    super(tokens);
	    assert(amI(tokens[0]));
	    if (tokens.length != 2) {
		throw new BadInsnException("wrong number of parameters: " +
					   tokens[0]);
	    }
	    outputs.add(new DataValue(tokens[1]));
	    inputs.add(new DataValue("eflags"));
	}

	final private static Collection names =
	    Arrays.asList("seta", "setae", "setb", "setbe", "setc", "sete",
			  "setg", "setge", "setl", "setle", "setna", "setnae",
			  "setnb", "setnbe", "setnc", "setne", "setng",
			  "setnge", "setnl", "setnle", "setno", "setnp",
			  "setns", "setnz", "seto", "setp", "setpe",
			  "setpo", "sets", "setz");

	static boolean amI(String op) {
	    return names.contains(op);
	}
    }

    /**
     * Move src to dst.
     */
    static class OpMov extends OpAssign {
	OpMov(String tokens[]) throws BadInsnException {
	    super(tokens);
	    assert(amI(tokens[0]));
	    if (tokens.length != 3) {
		throw new BadInsnException("wrong number of parameters: " +
					   tokens[0]);
	    }
	    inputs.add(new DataValue(tokens[1]));
	    outputs.add(new DataValue(tokens[2]));
	}

	final private static Collection names =
	    Arrays.asList("leal", "leaw", "lea",
			  "movzll", "movzwl", "movzbl",
			  "movzwl", "movzww", "movzwb",
			  "movzbl", "movzbw", "movzbb",
			  "movl", "movw", "movb");

	static boolean amI(String op) {
	    return names.contains(op);
	}
    }

    /**
     * Move (%esi) to (%edi), then increment or decrement depending on DF.
     */
    static class OpMovS extends OpMov {
	OpMovS(String tokens[]) throws BadInsnException {
	    super(tokens);
	    assert(amI(tokens[0]));
	    inputs.add(new DataValue("eflags"));
	    inputs.add(new DataValue("%esi"));
	    inputs.add(new DataValue("%edi"));
	    inputs.add(new DataValue("(%esi)"));
	    outputs.add(new DataValue("%esi"));
	    outputs.add(new DataValue("%edi"));
	    outputs.add(new DataValue("(%edi)"));
	}

	final private static Collection names =
	    Arrays.asList("movsll", "movslw", "movslb",
			  "movswl", "movsww", "movswb",
			  "movsbl", "movsbw", "movsbb");

	static boolean amI(String op) {
	    return names.contains(op);
	}
    }

    /**
     * An unconditional jump
     */
    static class OpJump extends OpUCControl {
	OpJump(String tokens[]) throws BadInsnException {
	    super(tokens);
	    assert(amI(tokens[0]));
	    if (tokens.length != 2) {
		throw new BadInsnException("wrong number of parameters: " +
					   tokens[0]);
	    }
	    label = tokens[1];
	}

	final private static Collection names =
	    Arrays.asList("jmp", "call");

	static boolean amI(String op) {
	    return names.contains(op);
	}

	// Call is slightly magical because we need to handle ret.
	public boolean isCall() {
	    return this.equals("call");
	}

	public String toString() {
	    return lineNum + "\t" + opcode + "\t" + label;
	}
    }

    /**
     * An unconditional return (no label)
     */
    static class OpRet extends OpUCControl {
	OpRet(String tokens[]) throws BadInsnException {
	    super(tokens);
	    assert(amI(tokens[0]));
	    if (tokens.length != 1) {
		throw new BadInsnException("wrong number of parameters: " +
					   tokens[0]);
	    }
	}

	static boolean amI(String op) {
	    return op.equals("ret");
	}

	public String toString() {
	    return lineNum + "\t" + opcode;
	}
    }

    /**
     * Conditional control transfer, e.g. je, jg
     */
    static class OpBranch extends OpControl {
	private boolean complained = false;

	OpBranch(String tokens[]) throws BadInsnException {
	    super(tokens);
	    label = tokens[1];
	    inputs = new LinkedList<DataValue>();
	    inputs.add(new DataValue("eflags"));
	}

	final private static Collection names =
	    Arrays.asList("ja", "jae", "jb", "jbe", "jc", "jcxz", "jecxz",
			  "je", "jg",
			  "jge", "jl", "jle", "jna", "jnae", "jnb", "jnbe",
			  "jnc", "jne", "jng", "jnge", "jnl", "jnle", "jno",
			  "jnp", "jns", "jnz", "jno", "jnp", "jnpe", "jpo",
			  "js", "jz");

	static boolean amI(String op) {
	    return names.contains(op);
	}

	public String toString() {
	    return lineNum + "\t" + opcode + "\t" + label;
	}

	public LocationSet execute(Configuration c) {
	    LocationSet data = c.getUntaintedData();

	    if (complained)
		return data;

	    for (DataValue in : inputs) {
		if (isTainted(data, in)) {
		    System.out.println("UNSAFE: " + toString());
		    Debug.println("\tuntainted locations: "
				  + data.toString());
		    complained = true;
		    return data;
		    // XXX should throw an exception that gets handled
		    // in Configuration by printing a trace of the
		    // offending path
		}
	    }
	    return data;
	}
    }

    /**
     * Anything that pops the stack
     */
    static class OpPop extends OpStack {
	OpPop(String tokens[]) throws BadInsnException {
	    super(tokens);
	    assert(amI(tokens[0]));
	    int expectedArgs;
	    if (tokens[0].equals("popfl")) {
		expectedArgs = 0;
		outputs.add(new DataValue("eflags"));
	    } else {
		expectedArgs = 1;
		outputs.add(new DataValue(tokens[1]));
	    }
	    if (tokens.length != expectedArgs + 1) {
		throw new BadInsnException("wrong number of parameters: " +
					   tokens[0]);
	    }
	}

	final private static Collection names =
	    Arrays.asList("popl", "popw", "popb",
			  "popfl");

	static boolean amI(String op) {
	    return names.contains(op);
	}
    }

    static class OpPush extends OpStack {
	OpPush(String tokens[]) throws BadInsnException {
	    super(tokens);
	    assert(amI(tokens[0]));
	    int expectedArgs;
	    if (tokens[0].equals("pushfl")) {
		expectedArgs = 0;
		inputs.add(new DataValue("eflags"));
	    } else {
		expectedArgs = 1;
		inputs.add(new DataValue(tokens[1]));
	    }
	    if (tokens.length != expectedArgs + 1) {
		throw new BadInsnException("wrong number of parameters: " +
					   tokens[0]);
	    }
	}

	final private static Collection names =
	    Arrays.asList("pushl", "pushw", "pushb",
			  "pushfl");

	static boolean amI(String op) {
	    return names.contains(op);
	}
    }    

    /**
     * Leave sets %esp and pops %ebp.
     */
    static class OpLeave extends OpStack {
	OpLeave(String tokens[]) throws BadInsnException {
	    super(tokens);
	    assert(amI(tokens[0]));
	    if (tokens.length != 1) {
		throw new BadInsnException("wrong number of parameters: " +
					   tokens[0]);
	    }
	    outputs.add(new DataValue("%ebp"));
	}

	static boolean amI(String op) {
	    return op.equals("leave");
	}
    }
}
