import java.util.*;

/**
 * Insn implementation for the IA32 architecture.
 */
abstract class InsnIA32 extends Insn {

    /**
     * This factory method returns an instance of a subclass of Insn,
     * corresponding to the appropriate kind of instruction.
     */
    public static InsnIA32 makeInsn(String tokens[], int lineNum)
	throws BadInsnException
    {
	try {
	    InsnIA32 insn;
	    if (OpAssignIO1.amI(tokens[0]))
		insn = new OpAssignIO1(tokens);
	    else if (OpAssignIO1NF.amI(tokens[0]))
		insn = new OpAssignIO1NF(tokens);
	    else if (OpAssignIO2.amI(tokens[0]))
		insn = new OpAssignIO2(tokens);
	    else if (OpAssignIO2F.amI(tokens[0]))
		insn = new OpAssignIO2F(tokens);
	    else if (OpCmp.amI(tokens[0]))
		insn = new OpCmp(tokens);
	    else if (OpSetCC.amI(tokens[0]))
		insn = new OpSetCC(tokens);
	    else if (OpMov.amI(tokens[0]))
		insn = new OpMov(tokens);
	    else if (OpMovS.amI(tokens[0]))
		insn = new OpMovS(tokens);
	    else if (OpJump.amI(tokens[0]))
		insn = new OpJump(tokens);
	    else if (OpRet.amI(tokens[0]))
		insn = new OpRet(tokens);
	    else if (OpBranch.amI(tokens[0]))
		insn = new OpBranch(tokens);
	    else if (OpPush.amI(tokens[0]))
		insn = new OpPush(tokens);
	    else if (OpPop.amI(tokens[0]))
		insn = new OpPop(tokens);
	    else if (OpLeave.amI(tokens[0]))
		insn = new OpLeave(tokens);
	    else
		throw new BadInsnException("unknown opcode " + tokens[0]);
	    insn.lineNum = lineNum;
	    return (insn);
	} catch (ArrayIndexOutOfBoundsException e) {
	    if (tokens[0] != null) {
		throw new BadInsnException("bad number of operands to " +
					   tokens[0]);
	    } else {
		throw new BadInsnException("no opcode");
	    }
	}
    }

    protected InsnIA32(String tokens[]) throws BadInsnException {
	super(tokens);
    }

    /**
     * Anything with inputs and outputs
     */
    static abstract class OpAssign extends InsnIA32 {
	OpAssign(String tokens[]) throws BadInsnException {
	    super(tokens);
	    inputs = new LinkedList<DataValue>();
	    outputs = new LinkedList<DataValue>();
	}

	public String toString() {
	    return lineNum + "\t" + opcode +
		"\tin: " + inputs + " out: " + outputs;
	}

	public LocationSet execute(Configuration c) {
	    LocationSet data = (LocationSet)c.getUntaintedData().clone();

	    checkMemoryAliasing(data);

	    // If any input is tainted, taint the output.
	    boolean taintOutputs = false;
	    for (DataValue in : inputs) {
		if (isTainted(data, in)) {
		    taintOutputs = true;
		    break;
		}
	    }
	    if (taintOutputs) {
		for (DataValue out : outputs)
		    data.remove(out);
	    } else {
		for (DataValue out : outputs)
		    data.add(out);
	    }

	    /*
	     * If an output operand was used to name a memory location,
	     * we need to re-taint any memory locations that it might
	     * have referred to.  This is true even if the new value
	     * being assigned to the pointer is a constant, since we know
	     * nothing about the new target of the pointer.
	     *
	     * XXX This is a bit expensive.
	     */
	    LinkedList<DataValue> toBeRemoved = new LinkedList<DataValue>();
	    for (DataValue d : data) {
		for (DataValue out : outputs) {
		    if (d.getDependencies().contains(out))
			toBeRemoved.add(d);
		}
	    }
	    for (DataValue d : toBeRemoved)
		data.remove(d);

	    return data;
	}
    }

    /**
     * Anything that transfers control
     */
    static abstract class OpControl extends InsnIA32 {
	OpControl(String tokens[]) throws BadInsnException {
	    super(tokens);
	}
    }

    /**
     * Unconditional control transfer, i.e. jmp and ret
     */
    static abstract class OpUCControl extends OpControl {
	OpUCControl(String tokens[]) throws BadInsnException {
	    super(tokens);
	}

	/** For unconditional control transfer, there's nothing to check. */
	public LocationSet execute(Configuration c) {
	    return c.getUntaintedData();
	}

	public boolean isUnconditionalControl() {
	    return true;
	}
    }

    /**
     * Anything that modifies the stack pointer
     */
    static abstract class OpStack extends OpAssign {
	OpStack(String tokens[]) throws BadInsnException {
	    super(tokens);
	    inputs.add(new DataValue("%esp"));
	    outputs.add(new DataValue("%esp"));
	}
    }

    /**
     * An assignment of the form OP IO, where operand and IO is both an
     * input and output operand.  Modifies eflags.
     */
    static class OpAssignIO1 extends OpAssign {
	OpAssignIO1(String tokens[]) throws BadInsnException {
	    super(tokens);
	    assert(amI(tokens[0]));
	    if (tokens.length != 2) {
		throw new BadInsnException("wrong number of parameters: " +
					   tokens[0]);
	    }
	    inputs.add(new DataValue(tokens[1]));
	    outputs.add(new DataValue(tokens[1]));
	    outputs.add(new DataValue("eflags"));
	}

	final private static Collection names =
	    Arrays.asList("decl", "decw", "decb",
			  "incl", "incw", "incb",
			  "negl", "negw", "negb");

	static boolean amI(String op) {
	    return names.contains(op);
	}
    }

    /**
     * An assignment of the form OP IO, where operand and IO is both an
     * input and output operand.  Does not modify eflags.
     */
    static class OpAssignIO1NF extends OpAssign {
	OpAssignIO1NF(String tokens[]) throws BadInsnException {
	    super(tokens);
	    assert(amI(tokens[0]));
	    if (tokens.length != 2) {
		throw new BadInsnException("wrong number of parameters: " +
					   tokens[0]);
	    }
	    inputs.add(new DataValue(tokens[1]));
	    outputs.add(new DataValue(tokens[1]));
	}

	final private static Collection names =
	    Arrays.asList("notl", "notw", "notb");

	static boolean amI(String op) {
	    return names.contains(op);
	}
    }

    /**
     * An assignment of the form OP IN, IO, where IN is an input
     * operand and IO is both an input and output operand.
     */
    static class OpAssignIO2 extends OpAssign {
	OpAssignIO2(String tokens[]) throws BadInsnException {
	    super(tokens);
	    assert(amI(tokens[0]));
	    if (tokens.length != 3) {
		throw new BadInsnException("wrong number of parameters: " +
					   tokens[0]);
	    }
	    inputs.add(new DataValue(tokens[1]));
	    inputs.add(new DataValue(tokens[2]));
	    outputs.add(new DataValue(tokens[2]));
	    outputs.add(new DataValue("eflags"));
	}

	/*
	 * XXX I assume that all of these set all 
	 * the interesting condition flags.  Only mostly true.
	 */
	final private static Collection names =
	    Arrays.asList("addl", "addw", "addb",
			  "andl", "andw", "andb",
			  "imull", "imulw", "imulb", // can also have 3 args
			  "orl",  "orw",  "orb",
			  "sall", "salw", "salb",
			  "sarl", "sarw", "sarb",
			  "sbbl", "sbbw", "sbbb",
			  "shll", "shlw", "shlb",
			  "shrl", "shlw", "shlb",
			  "subl", "subw", "subb",
			  "xorl", "xorw", "xorb");

	static boolean amI(String op) {
	    return names.contains(op);
	}
    }

    /**
     * Identical to OpAssignIO2, but eflags is also an input.
     */
    static class OpAssignIO2F extends OpAssignIO2 {
	OpAssignIO2F(String tokens[]) throws BadInsnException {
	    super(tokens);
	    assert(amI(tokens[0]));
	    inputs.add(new DataValue("eflags"));
	}

	final private static Collection names =
	    Arrays.asList("adcl", "adcw", "adcb");

	static boolean amI(String op) {
	    return names.contains(op);
	}
    }

    /**
     * Compare and set flags
     */
    static class OpCmp extends OpAssign {
	OpCmp(String tokens[]) throws BadInsnException {
	    super(tokens);
	    assert(amI(tokens[0]));
	    if (tokens.length != 3) {
		throw new BadInsnException("wrong number of parameters: " +
					   tokens[0]);
	    }
	    inputs.add(new DataValue(tokens[1]));
	    inputs.add(new DataValue(tokens[2]));
	    outputs.add(new DataValue("eflags"));
	}

	final private static Collection names =
	    Arrays.asList("cmpl", "cmpw", "cmpb",
			  "testl", "testw", "testb");

	static boolean amI(String op) {
	    return names.contains(op);
	}
    }

    /**
     * Set on condition code
     */
    static class OpSetCC extends OpAssign {
	OpSetCC(String tokens[]) throws BadInsnException {
	    super(tokens);
	    assert(amI(tokens[0]));
	    if (tokens.length != 2) {
		throw new BadInsnException("wrong number of parameters: " +
					   tokens[0]);
	    }
	    outputs.add(new DataValue(tokens[1]));
	    inputs.add(new DataValue("eflags"));
	}

	final private static Collection names =
	    Arrays.asList("seta", "setae", "setb", "setbe", "setc", "sete",
			  "setg", "setge", "setl", "setle", "setna", "setnae",
			  "setnb", "setnbe", "setnc", "setne", "setng",
			  "setnge", "setnl", "setnle", "setno", "setnp",
			  "setns", "setnz", "seto", "setp", "setpe",
			  "setpo", "sets", "setz");

	static boolean amI(String op) {
	    return names.contains(op);
	}
    }

    /**
     * Move src to dst.
     */
    static class OpMov extends OpAssign {
	OpMov(String tokens[]) throws BadInsnException {
	    super(tokens);
	    assert(amI(tokens[0]));
	    if (tokens.length != 3) {
		throw new BadInsnException("wrong number of parameters: " +
					   tokens[0]);
	    }
	    inputs.add(new DataValue(tokens[1]));
	    outputs.add(new DataValue(tokens[2]));
	}

	final private static Collection names =
	    Arrays.asList("leal", "leaw", "lea",
			  "movzll", "movzwl", "movzbl",
			  "movzwl", "movzww", "movzwb",
			  "movzbl", "movzbw", "movzbb",
			  "movl", "movw", "movb");

	static boolean amI(String op) {
	    return names.contains(op);
	}
    }

    /**
     * Move (%esi) to (%edi), then increment or decrement depending on DF.
     */
    static class OpMovS extends OpMov {
	OpMovS(String tokens[]) throws BadInsnException {
	    super(tokens);
	    assert(amI(tokens[0]));
	    inputs.add(new DataValue("eflags"));
	    inputs.add(new DataValue("%esi"));
	    inputs.add(new DataValue("%edi"));
	    inputs.add(new DataValue("(%esi)"));
	    outputs.add(new DataValue("%esi"));
	    outputs.add(new DataValue("%edi"));
	    outputs.add(new DataValue("(%edi)"));
	}

	final private static Collection names =
	    Arrays.asList("movsll", "movslw", "movslb",
			  "movswl", "movsww", "movswb",
			  "movsbl", "movsbw", "movsbb");

	static boolean amI(String op) {
	    return names.contains(op);
	}
    }

    /**
     * An unconditional jump
     */
    static class OpJump extends OpUCControl {
	OpJump(String tokens[]) throws BadInsnException {
	    super(tokens);
	    assert(amI(tokens[0]));
	    if (tokens.length != 2) {
		throw new BadInsnException("wrong number of parameters: " +
					   tokens[0]);
	    }
	    label = tokens[1];
	}

	final private static Collection names =
	    Arrays.asList("jmp", "call");

	static boolean amI(String op) {
	    return names.contains(op);
	}

	// Call is slightly magical because we need to handle ret.
	public boolean isCall() {
	    return this.equals("call");
	}

	public boolean hasLabelTarget() {
	    return true;
	}

	public String toString() {
	    return lineNum + "\t" + opcode + "\t" + label;
	}
    }

    /**
     * An unconditional return (no label)
     */
    static class OpRet extends OpUCControl {
	OpRet(String tokens[]) throws BadInsnException {
	    super(tokens);
	    assert(amI(tokens[0]));
	    if (tokens.length != 1) {
		throw new BadInsnException("wrong number of parameters: " +
					   tokens[0]);
	    }
	}

	static boolean amI(String op) {
	    return op.equals("ret");
	}

	public boolean isReturn() {
	    return true;
	}

	public String toString() {
	    return lineNum + "\t" + opcode;
	}
    }

    /**
     * Conditional control transfer, e.g. je, jg
     */
    static class OpBranch extends OpControl {
	private boolean complained = false;

	OpBranch(String tokens[]) throws BadInsnException {
	    super(tokens);
	    label = tokens[1];
	    inputs = new LinkedList<DataValue>();
	    inputs.add(new DataValue("eflags"));
	}

	final private static Collection names =
	    Arrays.asList("ja", "jae", "jb", "jbe", "jc", "jcxz", "jecxz",
			  "je", "jg",
			  "jge", "jl", "jle", "jna", "jnae", "jnb", "jnbe",
			  "jnc", "jne", "jng", "jnge", "jnl", "jnle", "jno",
			  "jnp", "jns", "jnz", "jno", "jnp", "jnpe", "jpo",
			  "js", "jz");

	static boolean amI(String op) {
	    return names.contains(op);
	}

	public boolean hasLabelTarget() {
	    return true;
	}

	public String toString() {
	    return lineNum + "\t" + opcode + "\t" + label;
	}

	public LocationSet execute(Configuration c) {
	    LocationSet data = c.getUntaintedData();

	    if (complained)
		return data;

	    for (DataValue in : inputs) {
		if (isTainted(data, in)) {
		    System.out.println("UNSAFE: " + toString());
		    Debug.println("\tuntainted locations: "
				  + data.toString());
		    complained = true;
		    return data;
		    // XXX should throw an exception that gets handled
		    // in Configuration by printing a trace of the
		    // offending path
		}
	    }
	    return data;
	}
    }

    /**
     * Anything that pops the stack
     */
    static class OpPop extends OpStack {
	OpPop(String tokens[]) throws BadInsnException {
	    super(tokens);
	    assert(amI(tokens[0]));
	    int expectedArgs;
	    if (tokens[0].equals("popfl")) {
		expectedArgs = 0;
		outputs.add(new DataValue("eflags"));
	    } else {
		expectedArgs = 1;
		outputs.add(new DataValue(tokens[1]));
	    }
	    if (tokens.length != expectedArgs + 1) {
		throw new BadInsnException("wrong number of parameters: " +
					   tokens[0]);
	    }
	}

	final private static Collection names =
	    Arrays.asList("popl", "popw", "popb",
			  "popfl");

	static boolean amI(String op) {
	    return names.contains(op);
	}
    }

    static class OpPush extends OpStack {
	OpPush(String tokens[]) throws BadInsnException {
	    super(tokens);
	    assert(amI(tokens[0]));
	    int expectedArgs;
	    if (tokens[0].equals("pushfl")) {
		expectedArgs = 0;
		inputs.add(new DataValue("eflags"));
	    } else {
		expectedArgs = 1;
		inputs.add(new DataValue(tokens[1]));
	    }
	    if (tokens.length != expectedArgs + 1) {
		throw new BadInsnException("wrong number of parameters: " +
					   tokens[0]);
	    }
	}

	final private static Collection names =
	    Arrays.asList("pushl", "pushw", "pushb",
			  "pushfl");

	static boolean amI(String op) {
	    return names.contains(op);
	}
    }    

    /**
     * Leave sets %esp and pops %ebp.
     */
    static class OpLeave extends OpStack {
	OpLeave(String tokens[]) throws BadInsnException {
	    super(tokens);
	    assert(amI(tokens[0]));
	    if (tokens.length != 1) {
		throw new BadInsnException("wrong number of parameters: " +
					   tokens[0]);
	    }
	    outputs.add(new DataValue("%ebp"));
	}

	static boolean amI(String op) {
	    return op.equals("leave");
	}
    }
}
